/*
 ===========================================================================
 Seek Patch for Samsung TV, developed by Drin
 ===========================================================================
*/

#define _GNU_SOURCE
#define VERSION "0.2"
#define BUILD_DATE "Tue, 15 Mar 2011 12:22:46 +0200"

#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/mman.h>
#include "seekpatch.h"

#ifdef DEBUG
static FILE *hlog;
#endif

static char * parseValue(char *s, char *name)
{
	char *arg, *arg_val;
	int arg_size;

	while(s) {
		arg = strstr(s, ",");
		if(arg == NULL) {
			arg = s;
			s = NULL;
		} else {
			arg_size = arg - s;
			arg = strndup(s, arg_size);
			s += arg_size + 1;
		}
		arg_val = strstr(arg, "=");
		if(arg_val != NULL) {
			arg_size = arg_val - arg;
			if(strcasecmp(strndup(arg, arg_size), name) == 0) {
				arg_val++;
				return arg_val;
			}
		}
	}
	return NULL;
}

static float parseValueF(char *s, char *name, float defaultVal)
{
	s = parseValue(s, name);
	if(s == NULL) return defaultVal;
	float result;
	return sscanf(s, "%f", &result) ? result : defaultVal;
}

static void cacheflush(void *start, size_t size)
{
	#ifdef __ARM_EABI__
   asm("mov r0, %0\n"
       "mov r1, %1\n"
       "mov r2, #0\n"
       "mov r7, #0xf0000\n"
       "add r7, #0x02\n"
       "swi #0\n" ::
       "g"(start), "g"((long)start+size) :
       "r0", "r1", "r2", "r7");
	#endif
}

/*
Searching insruction by mask: LDR[condflag] R[x], ??
from [startOffset] to [startOffset] + [size]
if condflag < 0 search any LDR?
if x < 0 search any (R0-R15)

OPSTRUCT: CONDFLAG(4bits)OPCODE(12bits)REGNUM(4bits)OFFSET(12bits)
*/
static unsigned __ARM_search_LDR_R(int condflag, int x, unsigned startOffset, unsigned size)
{
	if(condflag > __ARM_CONDITION__NV) return 0;
	if(x > 0xF) return 0;

	unsigned *ptr;

	for(size+=startOffset; startOffset < size; startOffset+=4)
	{
		ptr = (unsigned *) startOffset;
		if( (condflag >= 0) && (condflag != (*(ptr) >> 28)) ) continue; // not ?[condflag]
		if( ((*(ptr) & 0x0FFFFFFF) >> 16) != 0x59F ) continue; // not LDR R?
		if( (x >= 0) && (x != ((*(ptr) & 0xF000) >> 12)) ) continue; // not LDR R[x]
		return (unsigned) ptr;
	}
	return 0;
}

static float* __ARM_search_LDR_R_VAL(int condflag, int x, unsigned startOffset, unsigned size, float val)
{
	unsigned *ptr;
	float *ptrSeg;

	while( (size > 0) && (ptr = (unsigned *) __ARM_search_LDR_R(condflag, x, startOffset, size)) ) {
		ptrSeg = (float *) ((unsigned) ptr + 8 + (*(ptr) & 0xFFF));
		if( *(ptrSeg) == val) return ptrSeg;
		size -= (unsigned) ptr + 4 - startOffset;
		startOffset = (unsigned) ptr + 4;
	}
	return NULL;
}

static void patchFloat(float *addr, float val)
{
	const unsigned page = ((unsigned) addr) & ~0x0FFF;
	mprotect((void *) page, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	*(addr) = val;
	mprotect((void *) page, 8192, PROT_READ | PROT_EXEC);
	cacheflush(addr, 4);
}

int patch_main(char *path, const char *udn __attribute__ ((unused)))
{
	void *myself;

#ifdef DEBUG
	hlog = fopen( "/dtv/seekmovie.log", "a+" );
	fprintf(hlog, "Seek Patch "VERSION" for Samsung TV, developed by Drin, builded "BUILD_DATE"\n\n");
	fflush(hlog);
#endif

	if ( (myself = dlopen(NULL,RTLD_NOW | RTLD_GLOBAL)) == NULL )
	{
#ifdef DEBUG
		fprintf(hlog, "Can't dlopen myself\n");
		fclose(hlog);
#endif
		return 0;
	} //endif

	const float argUp = parseValueF(path, "UP", 0);
	const float argDown = parseValueF(path, "DOWN", 0);
	const float argRight = parseValueF(path, "RIGHT", 0);
	const float argLeft = parseValueF(path, "LEFT", 0);
#ifdef DEBUG
	fprintf(hlog, "argUp: %f, argDown: %f, argRight: %f, argLeft: %f\n", argUp, argDown, argRight, argLeft);
	fflush(hlog);
#endif
	if( (argUp == 0) && (argDown == 0) && (argRight == 0) && (argLeft == 0) )
	{
#ifdef DEBUG
		fprintf(hlog, "NOTHING TO PATCH!\nUSAGE EXAMPLE: injectso `pidof exeDSP` /SamyGO/seekpatch.so patch_main UP=120,DOWN=-180,RIGHT=30,LEFT=-60\n");
		fclose(hlog);
#endif
		return 0;
	} //endif

	const unsigned *CWProVideoPlayer__ProcessDirectionKey = (unsigned *) dlsym(myself,"_ZN16CWProVideoPlayer19ProcessDirectionKeyEPK7PTEvent");
	dlclose(myself);
	if ( CWProVideoPlayer__ProcessDirectionKey == NULL )
	{
#ifdef DEBUG
		fprintf(hlog, "Can't locate CWProVideoPlayer::ProcessDirectionKey\n");
		fclose(hlog);
#endif
		return 0;
	} //endif

	float *pSeekUp = __ARM_search_LDR_R_VAL(__ARM_CONDITION__AL, 4, (unsigned) CWProVideoPlayer__ProcessDirectionKey, 1024, 214748368.);
	float *pSeekDown = __ARM_search_LDR_R_VAL(__ARM_CONDITION__AL, 4, (unsigned) CWProVideoPlayer__ProcessDirectionKey, 1024, -1073741824.);
	float *pSeekRight = __ARM_search_LDR_R_VAL(__ARM_CONDITION__NE, -1, (unsigned) CWProVideoPlayer__ProcessDirectionKey, 1024, 10.);
	float *pSeekLeft = __ARM_search_LDR_R_VAL(__ARM_CONDITION__NE, -1, (unsigned) CWProVideoPlayer__ProcessDirectionKey, 1024, -10.);
	if( (pSeekUp == NULL) || (pSeekUp == NULL) || (pSeekRight == NULL) )
	{
#ifdef DEBUG
		fprintf(hlog, "ERROR! Some offsets not found....\n");
		fclose(hlog);
#endif
		return 0;
	} //endif

	// Patching
	if(argUp != 0)
	{
#ifdef DEBUG
		fprintf(hlog, "SeekUp value offset [0x%.8X]: going change %0.2f -> %0.2f\n", (unsigned) pSeekUp, *(pSeekUp), argUp);
		fflush(hlog);
#endif
		patchFloat(pSeekUp, argUp);
	} //endif
	if(argDown != 0)
	{
#ifdef DEBUG
		fprintf(hlog, "SeekDown value offset [0x%.8X]: going change %0.2f -> %0.2f\n", (unsigned) pSeekDown, *(pSeekDown), argDown);
		fflush(hlog);
#endif
		patchFloat(pSeekDown, argDown);
	} //endif
	if(argRight != 0)
	{
#ifdef DEBUG
		fprintf(hlog, "SeekRight value offset [0x%.8X]: going change %0.2f -> %0.2f\n", (unsigned) pSeekRight, *(pSeekRight), argRight);
		fflush(hlog);
#endif
		patchFloat(pSeekRight, argRight);
	} //endif
	if(argLeft != 0)
	{
#ifdef DEBUG
		fprintf(hlog, "SeekLeft value offset [0x%.8X]: going change %0.2f -> %0.2f\n", (unsigned) pSeekLeft, *(pSeekLeft), argLeft);
		fflush(hlog);
#endif
		patchFloat(pSeekLeft, argLeft);
	} //endif

#ifdef DEBUG
	fprintf(hlog, "DONE!\n");
	fclose(hlog);
#endif
	return 1;
}
